/*
    Xspider, an X11 spider friend
    Copyright (C) 2022-2023  alicia@ion.nu

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <X11/Xlib.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkx.h>

extern GdkPixbuf* rotatescale(GdkPixbuf* src, double angle, int size);

const double rpad=-4; // Extra rotation to make the image match the direction (in radians)
static GdkRectangle* geometries;
static int nmonitors;

struct spider
{
  int size;
  double r;
  double pos[2];
  char walking;
  GtkWidget* imgwidget;
  int pad;
  char drawn;
};
struct spider spider={40,0};
GdkPixbuf* img;
Display* xdisplay;
Window xroot;

int findscreen(int x, int y)
{
  int i;
  for(i=0; i<nmonitors; i++)
  {
    if(x>=geometries[i].x &&
       y>=geometries[i].y &&
       x<geometries[i].x+geometries[i].width &&
       y<geometries[i].y+geometries[i].height)
    {
      return i;
    }
  }
  return -1;
}

gboolean drawspider(gpointer ptr)
{
  struct spider* s=&spider;
  char action=random()%18;
  // Simplify r for the edge avoidance conditions below
  s->r=fmod(s->r, M_PI*2);
  if(s->r<0){s->r+=M_PI*2;}

  // New edge detection using all monitors
  int x=s->pos[0]+s->size/2; // Use these coordinates to figure out if there's another monitor to which could walk
  int y=s->pos[1]+s->size/2;
  int currentscreen=findscreen(x, y); // First find the base monitor
  // r starts facing left and goes clockwise
  if(s->r<M_PI/2 || s->r>M_PI*1.5){x-=s->size/2;}
  if(s->r>M_PI/2 && s->r<M_PI*1.5){x+=s->size/2;}
  if(s->r<M_PI){y-=s->size/2;}
  if(s->r>M_PI){y+=s->size/2;}
  char haveneighborscreen=(findscreen(x, y)>=0);
  GdkRectangle* g=&geometries[currentscreen]; // Figure out if we have other monitors to walk over to
  if(!haveneighborscreen)
  {
    // Avoid edges, preferably by turning away from the edge
    if(s->pos[0]<=g->x && (s->r<M_PI/2 || s->r>M_PI*1.5)){action=s->r<M_PI?0:1;}
    if(s->pos[1]<=g->y && s->r<M_PI){action=s->r<M_PI/2?1:0;}
    if(s->pos[0]>=g->x+g->width-s->size && s->r>M_PI/2 && s->r<M_PI*1.5){action=s->r<M_PI?1:0;}
    if(s->pos[1]>=g->y+g->height-s->size && s->r>M_PI){action=s->r<M_PI*1.5?1:0;}
  }

  switch(action)
  {
  case 0: s->r+=M_PI/18; break; // Turn
  case 1: s->r-=M_PI/18; break;
  // For a more realistic jumping spider walk, randomly start and stop walking
  case 2: s->walking=1; break;
  case 3:
  case 4: s->walking=0; break;
  case 5: // Pounce, if possible
    if(xdisplay)
    {
      int x, y;
      // Unwanted return values, but the function demands them
      union{Window a;int b;unsigned int c;} ret;
      // Get mouse position
      XQueryPointer(xdisplay, xroot, &ret.a, &ret.a, &x, &y, &ret.b, &ret.b, &ret.c);
      int xdist=(s->pos[0]+s->size/2)-x;
      int ydist=(s->pos[1]+s->size/2)-y;
      int dist=sqrt(xdist*xdist+ydist*ydist);
      if(dist>s->size*0.75 && dist<s->size*2)
      {
        double target=atan2(ydist, xdist);
        if(target<0){target+=M_PI*2;}
//printf("Within distance for pounce. Angle: %f\n", target);
        if(fabs(target-s->r)<M_PI/4) // 180°/4=45 (+/-, so a 90° cone of vision)
        {
          s->r=target;
          // Go to cursor, but not quite centered on it
          s->pos[0]-=cos(s->r)*(dist-s->size/3);
          s->pos[1]-=sin(s->r)*(dist-s->size/3);
          s->walking=0; // Stop for a bit
          s->drawn=0; // Force redraw
        }
      }
    }
    break;
  }
  if(s->walking)
  {
    s->pos[0]-=cos(s->r);
    s->pos[1]-=sin(s->r);
    if(!haveneighborscreen)
    {
      if(s->pos[0]<g->x){s->pos[0]=g->x;}
      if(s->pos[1]<g->y){s->pos[1]=g->y;}
      if(s->pos[0]>g->x+g->width-s->size){s->pos[0]=g->x+g->width-s->size;}
      if(s->pos[1]>g->y+g->height-s->size){s->pos[1]=g->y+g->height-s->size;}
    }
  }
  GdkWindow* window=ptr;
  gdk_window_move(window, s->pos[0], s->pos[1]);
  if(action>1 && s->drawn){return G_SOURCE_CONTINUE;} // No rotate = no need to update image/window shape
  s->drawn=1; // Gotta draw it on startup at least

  GdkPixbuf* rotated=rotatescale(img, s->r+rpad, s->size);
  gtk_image_set_from_pixbuf(GTK_IMAGE(s->imgwidget), rotated);
  guchar* spider=gdk_pixbuf_get_pixels(rotated);
  int rowstride=gdk_pixbuf_get_rowstride(rotated);
  int pstride=gdk_pixbuf_get_n_channels(rotated);

  cairo_rectangle_int_t* rects=0;
  int count=0;
  for(y=0; y<s->size; y++)
  {
    for(x=0; x<s->size; x++)
    {
      if(spider[y*rowstride+x*pstride+3]<64){continue;}
      count++;
      rects=realloc(rects, sizeof(cairo_rectangle_int_t)*count);
      rects[count-1].x=x;
      rects[count-1].y=y;
      rects[count-1].width=1;
      rects[count-1].height=1;
    }
  }
  cairo_region_t* region=cairo_region_create_rectangles(rects, count);
  gdk_window_shape_combine_region(window, region, 0, 0);
  free(rects);
  s->pad=1-s->pad;
  // Resizing updates the window
  gdk_window_resize(window, s->size, s->size+s->pad);
  cairo_region_destroy(region);
  g_object_unref(rotated);
  return G_SOURCE_CONTINUE;
}

int main(int argc, char** argv)
{
  srandom(time(0));
  gtk_init(0,0);
  GtkWidget* window=gtk_window_new(GTK_WINDOW_POPUP);
  gtk_window_set_decorated(GTK_WINDOW(window), 0);
  gtk_window_resize(GTK_WINDOW(window), spider.size, spider.size);
  gtk_window_set_keep_above(GTK_WINDOW(window), 1);
  gtk_widget_realize(window);
  GdkWindow* gwindow=gtk_widget_get_window(window);
  if(!gwindow){printf("No GdkWindow yet\n");return 1;}
  gdk_window_set_shadow_width(gwindow, 0, 0, 0, 0);
  gdk_window_focus(gwindow, 0);
  gdk_window_set_decorations(gwindow, 0);
  gdk_window_set_type_hint(gwindow, GDK_WINDOW_TYPE_HINT_TOOLTIP);
  gdk_window_set_opacity(gwindow, 0.99); // Weird little trick to get rid of the window shadows
  GdkDisplay* display=gdk_display_get_default();

  xdisplay=gdk_x11_display_get_xdisplay(display);
  int xscreen=XDefaultScreen(xdisplay);
  xroot=RootWindow(xdisplay, xscreen);

  nmonitors=gdk_display_get_n_monitors(display);
  GdkRectangle geom[nmonitors];
  int i;
  for(i=0; i<nmonitors; i++)
  {
    GdkMonitor* monitor=gdk_display_get_monitor(display, i);
    gdk_monitor_get_geometry(monitor, &geom[i]);
//    printf("Monitor %i: %ix%i+%i+%i\n", i, geom[i].width, geom[i].height, geom[i].x, geom[i].y);
  }
  geometries=geom;

  // Start with random position and direction
  int startscreen=random()*nmonitors/RAND_MAX;
  spider.r=(double)random()*M_PI*2/RAND_MAX;//360;
  spider.pos[0]=random()%(geom[startscreen].width-spider.size);
  spider.pos[1]=random()%(geom[startscreen].height-spider.size);
  spider.pos[0]+=geom[startscreen].x;
  spider.pos[1]+=geom[startscreen].y;
  // Or use specified values. TODO: Flags instead of positionals. Include spidersize
  if(argc>1){spider.pos[0]=atoi(argv[1]);}
  if(argc>2){spider.pos[1]=atoi(argv[2]);}
  if(argc>3){spider.r=strtod(argv[3], 0);}
  // Load image
  img=gdk_pixbuf_new_from_resource("/org/xspider/spider.png", 0);
  spider.imgwidget=gtk_image_new();

  drawspider(gwindow);
  gtk_widget_show(window);
  g_timeout_add(100, drawspider, gwindow);
  gtk_container_add(GTK_CONTAINER(window), spider.imgwidget);
  gtk_widget_show(spider.imgwidget);
  gtk_main();
}
